import pmpu_generated_files as cfg
import os
import argparse
import numpy as np
import pandas as pd
import random
import itertools


# This function generates all the possible configurations (states) of the swimmer based on all the possible lengths of its arm
# and the radii of its left sphere
def generate_states(possible_radii, possible_arm_lengths):
    states = [list(p) for p in itertools.product(possible_radii, possible_arm_lengths)]
    return states


# This function generates all the actions that the swimmer can takes based on all the possible lengths of its arm
# and the radii of its left sphere
def generate_actions(possible_radii, possible_arm_lengths):
    action1 = "inflate_left_sphere"
    action2 = "deflate_left_sphere"
    action3 = "retract_arm"
    action4 = "extend_arm"
    possible_radii_diff = []
    possible_relative_displacements = []
    for i in range(0, len(possible_radii)):
        for j in range(0, len(possible_radii)):
            possible_radii_diff.append(abs(possible_radii[i]-possible_radii[j]))
    possible_radii_diff = list(set(possible_radii_diff))
    possible_radii_diff.remove(0)
    for i in range(0, len(possible_arm_lengths)):
        for j in range(0, len(possible_arm_lengths)):
            possible_relative_displacements.append(abs(possible_arm_lengths[i]-possible_arm_lengths[j]))
    possible_relative_displacements = list(set(possible_relative_displacements))
    possible_relative_displacements.remove(0)
    actions_inflate = [[action1, possible_radii_diff[i]] for i in range(0,len(possible_radii_diff))]
    actions_deflate = [[action2, -possible_radii_diff[i]] for i in range(0,len(possible_radii_diff))]
    actions_arm = [action3, action4]
    list_actions_arm = [actions_arm, possible_relative_displacements]
    all_actions_radii = actions_inflate+actions_deflate
    all_actions_arm = [list(p) for p in itertools.product(*list_actions_arm)]
    actions = all_actions_radii+all_actions_arm
    return actions


# Generating automatic simulations for the 2D and 3D PMPY swimmer based on the state of the swimmer and the action to take
# The argumet 'Dim' specifies the dimention of the PMMY swimmer
def generate_simulation_pmpu(state, action, Dim) :
    os.makedirs('../sw/pmpu_'+str(Dim)+'D', exist_ok=True)  
    cfg.write_cfg('../sw/pmpu_'+str(Dim)+'D', Dim)
    cfg.write_geo(state, '../sw/pmpu_'+str(Dim)+'D', Dim) 
    cfg.write_json(action, state, '../sw/pmpu_'+str(Dim)+'D', Dim)
    cfg.write_fluid_remesh('..', Dim)
    os.system('mpirun -np 10 python3 ../fluid-remesh_for_pmpu.py')
    data = pd.read_csv("../Reinf_Learn_PMPY/fluid.measures.csv")
    data.columns = data.columns.str.strip()
    sph_left_Xposition = list(data['Quantities_body_SphereLeft.mass_center_0'])
    sph_right_Xposition = list(data['Quantities_body_SphereRight.mass_center_0'])
    time = list(data['time'])
    Delta_right = sph_right_Xposition[len(time)-2]-sph_right_Xposition[0]
    Delta_left = sph_left_Xposition[len(time)-2]-sph_left_Xposition[0]
    reward = (Delta_right+Delta_left)/2
    return reward


# This function returns the new configuration of the swwimmer after taking a certain action
def new_state(state, action):
    if action[0] == "retract_arm" :
        s = [state[0], state[1]-action[1]]
    elif action[0] == "extend_arm" :
        s = [state[0], state[1]+action[1]]
    elif action[0] == "deflate_left_sphere" or action[0] == "inflate_left_sphere":
        s = [state[0]+action[1], state[1]]
    return s


# For a geven state, this function returns all the possible actions that the swimmer can take
def possible_actions(state, all_states, all_actions):
    all_states_float = [str(all_states[i]) for i in range(0, len(all_states))]
    state_float = [float(state[0]), float(state[1])]
    actions = []
    for action in all_actions :
        if new_state(state, action) in all_states :
            actions.append(action)
    return actions


# Return the index of dictinary value
def get_key(val, dict):
    for key, value in dict.items():
        if val == value:
            return key


# This function returns the indexes of all the possible actions that can be taken in a certain state
def get_index_of_possible_actions(state, states, actions, actions_space):
    indexes = []
    for action in possible_actions(state, states, actions):
        indexes.append(get_key(action, actions_space))
    return indexes


# This function returns the argmax of the Q-table over a specific set of actions
def get_max_Q_table_index(possible_action_indexes, state_index, Q_table):
    Q_table_possible_actions = {possible_action_indexes[i] : Q_table[state_index, possible_action_indexes[i]] for i in range(0, len(possible_action_indexes))}
    Q_table_possible_actions_list = [Q_table[state_index, possible_action_indexes[i]] for i in range(0, len(possible_action_indexes))]
    Q_table_max = max(Q_table_possible_actions_list)
    index_max = get_key(Q_table_max, Q_table_possible_actions)
    return index_max


# This function returns the argmax of the two Q-tables (for Double Q-learning) over a specific set of actions
def get_max_Q_tables_index(indexes, state_index, Q_a_table, Q_b_table):
    Q_a = {indexes[0] : Q_a_table[state_index, indexes[0]], indexes[1] : Q_a_table[state_index, indexes[1]]}
    Q_b = {indexes[0] : Q_b_table[state_index, indexes[0]], indexes[1] : Q_b_table[state_index, indexes[1]]}
    lis_a = [Q_a[indexes[0]], Q_a[indexes[1]]]
    lis_b = [Q_b[indexes[0]], Q_b[indexes[1]]]
    max_val = max(Q_a_table[state_index, indexes[0]], Q_a_table[state_index, indexes[1]], Q_b_table[state_index, indexes[0]], Q_b_table[state_index, indexes[1]])    
    if max_val in lis_a :
        index_max = get_key(max_val, Q_a)
    elif max_val in lis_b :
        index_max = get_key(max_val, Q_b)
    return index_max


# Generating a dictionary for a list to index each element (to be used for the Q-matrix)
def generate_dict(lst):
    lst_space = {}
    for i in range(0, len(lst)):
        lst_space[i] = lst[i]
    return lst_space


# The function determine the optimal stroke for the PMPY model depending on the values of the possible radii of its sphere and the possible lengths of its arm.
# We note that this optimal stroke was figured out from the expression of the PMPY displacement given in Averon paper.
def optimal_strategy(possible_radii, possible_arm_lengths):
    state_1 = [min(possible_radii), min(possible_arm_lengths)]
    state_2 = [max(possible_radii), min(possible_arm_lengths)]
    state_3 = [max(possible_radii), max(possible_arm_lengths)]
    state_4 = [min(possible_radii), max(possible_arm_lengths)]
    optimal_strategy_pmpu = [state_1, state_2, state_3, state_4]
    return optimal_strategy_pmpu


# This function determines the required number of learning steps to learn the optimal swimming strategy...
# Once the swimmer repeats this optimal strategy two times, it is considered to learn it.
def learning_steps_required(optimal_strategy, states_list):
    required_number = 0
    for i in range(0,len(states_list)-2*len(optimal_strategy)):
        if states_list[i] in optimal_strategy :
            index = optimal_strategy.index(states_list[i])
            opt_strategy = optimal_strategy[index%len(optimal_strategy):(index%len(optimal_strategy))+len(optimal_strategy)]
            for j in range(0,index%len(optimal_strategy)):
                opt_strategy.append(optimal_strategy[j])
            if states_list[i:i+2*len(optimal_strategy)] == 2*opt_strategy :
                required_number = i
                break
    return required_number


# This function checks if the rewards have been computed or not. It returns a matrix and a path.
# The matrix takes the rewards if they have been computed. If not, the rewards that will be computed
# will be saved in the returned path
def check_stored_rewards(possible_radii, possible_arm_lengths, n, m, Dim):
    st = 'radii='
    possible_radii.sort()
    st+= str(possible_radii)
    possible_arm_lengths.sort()
    st += '_arm_lengths='+str(possible_arm_lengths)
    st = st.replace(" ", "")
    path = "../Reinf_Learn_PMPY/rewards_pmpu_"+str(Dim)+"D_"+st+".csv"
    if os.path.isfile(path):
        reader = pd.read_csv(path)
        stored_rewards = np.array(reader).astype("float")[:, 1:]
    else :
        stored_rewards = np.zeros((n, m))
    return stored_rewards, path


# Launching the Q-learning process 
def Qlearning(possible_radii, possible_arm_lengths, N_learn, eps, gamma, alpha, Dim):
    actions = generate_actions(possible_radii, possible_arm_lengths)
    states = generate_states(possible_radii, possible_arm_lengths)
    n = len(states)
    m = len(actions)
    actions_space = generate_dict(actions)  # Generating a dictionary for the set of actions (for indexation)
    states_space = generate_dict(states)    # Generating a dictionary for the set of actions
    state = random.choice(states)           # Initial state is chosen randomly
    cumul_reward = 0                        # Initialisation of the cumulative reward 
    cumul_reward_list = [0]                 # Initialization of the list that takes the cumulative rewards
    actions_list = []                       # Initialization of the list that takes the chosen actions over the Q-learning process
    states_list = [state]                   # Initialization of the list that takes the states over the Q-learning process
    Q_table = np.zeros((n, m))              # Initialization of the Q-table
    stored_rewards, path = check_stored_rewards(possible_radii, possible_arm_lengths, n, m, Dim)
    # Learning process
    for i in range(0, N_learn):
        state_index = get_key(state, states_space)     # Get the index of the state to be used to access Q-table
        # Epsilon-greedy scheme
        if random.uniform(0,1) < eps : 
            action = random.choice(possible_actions(state, states, actions))
            action_index = get_key(action, actions_space)   
        else :
            possible_action_indexes = get_index_of_possible_actions(state, states, actions, actions_space)
            action_index = get_max_Q_table_index(possible_action_indexes, state_index, Q_table)
            action = actions_space[action_index]
        # We check if the reward of the pair action-state is stored, otherwise we perform a simulation and store the reward to avoid repeating simulations
        if stored_rewards[state_index, action_index] == 0:
            reward = generate_simulation_pmpu(state, action, Dim)
            New_state = new_state(state, action)
            stored_rewards[state_index, action_index] = reward
            pd.DataFrame(stored_rewards).to_csv(path)
        else :
            reward = stored_rewards[state_index, action_index]
            New_state = new_state(state, action)
        # print("state = ", state)
        New_state_index = get_key(New_state, states_space)     # get the index of the new state
        New_indexes = get_index_of_possible_actions(New_state, states, actions, actions_space)
        Next_state_Q_values = [Q_table[New_state_index, j] for j in New_indexes]
        max_value = max(Next_state_Q_values)
        # Updating the Q-table
        Q_table[state_index, action_index] = Q_table[state_index, action_index] + alpha*(reward + gamma*max_value - Q_table[state_index, action_index])
        cumul_reward += reward                  # Updating the cumulative reward
        state = New_state                       # Updating the state
        cumul_reward_list.append(cumul_reward)
        actions_list.append(action)
        states_list.append(New_state)
    return cumul_reward_list, states_list


# Launching the double Q-learning process 
def Double_Qlearning(possible_radii, possible_arm_lengths, N_learn, eps, gamma, alpha, Dim):
    actions = generate_actions(possible_radii, possible_arm_lengths)
    states = generate_states(possible_radii, possible_arm_lengths)
    n = len(states)
    m = len(actions)
    actions_space = generate_dict(actions)      # Generating a dictionary for the set of actions (for indexation)
    states_space = generate_dict(states)        # Generating a dictionary for the set of actions
    state = random.choice(states)               # Initial state is chosen randomly
    cumul_reward = 0                            # Initialisation of the cumulative reward 
    cumul_reward_list = [0]                     # Initialization of the list that takes the cumulative rewards
    actions_list = []                           # Initialization of the list that takes the chosen actions over the Q-learning process
    states_list = [state]                       # Initialization of the list that takes the states over the Q-learning process
    Q_a_table = np.zeros((n, m))
    Q_b_table = np.zeros((n, m))
    Q_tables = ["Q_a_table", "Q_b_table"]
    stored_rewards, path = check_stored_rewards(possible_radii, possible_arm_lengths, n, m, Dim)
    # Learning process
    for i in range(0, N_learn):
        state_index = get_key(state, states_space)     # Get the index of the state to be used to access Q-table
        # Epsilon-greedy scheme
        if random.uniform(0,1) < eps : 
            action = random.choice(possible_actions(state, states, actions))
            action_index = get_key(action, actions_space)
        else :
            possible_action_indexes = get_index_of_possible_actions(state, states, actions, actions_space)
            action_index = get_max_Q_tables_index(possible_action_indexes, state_index, Q_a_table, Q_b_table)
            action = actions_space[action_index]
        # We check if the reward of the pair action-state is stored, otherwise we perform a simulation and store the reward to avoid repeating simulations
        if stored_rewards[state_index, action_index] == 0:
            reward = generate_simulation_pmpu(state, action, Dim)
            New_state = new_state(state, action)
            stored_rewards[state_index, action_index] = reward
            pd.DataFrame(stored_rewards).to_csv(path)
        else :
            reward = stored_rewards[state_index, action_index]
            New_state = new_state(state, action)
        print("state = ", state)
        New_state_index = get_key(New_state, states_space)     # get the index of the new state
        New_indexes = get_index_of_possible_actions(New_state, states, actions, actions_space)
        # Updating one of the Q-tables (random choice)
        updated = random.choice(Q_tables)
        if updated == "Q_a_table" :
            action_a_index = get_max_Q_table_index(New_indexes, New_state_index, Q_a_table)
            Q_a_table[state_index, action_index] = Q_a_table[state_index, action_index] + alpha*(reward + gamma*Q_b_table[New_state_index, action_a_index] - Q_a_table[state_index, action_index])
        elif updated == "Q_b_table" :
            action_b_index = get_max_Q_table_index(New_indexes, New_state_index, Q_b_table)
            Q_b_table[state_index, action_index] = Q_b_table[state_index, action_index] + alpha*(reward + gamma*Q_a_table[New_state_index, action_b_index] - Q_b_table[state_index, action_index])
        cumul_reward += reward                  # Uptating the cumulative reward
        state = New_state                       # Updating the state
        cumul_reward_list.append(cumul_reward)
        actions_list.append(action)
        states_list.append(New_state) 
    return cumul_reward_list, states_list


